#include "string.h"
#include "avr/pgmspace.h"
#include "util/crc16.h"
#include "util/atomic.h"
#include "util/twi.h"

#include "cdelay.h"
#include "ddrpin.h"

#include "eep_i2c.h"

#include "eep_i2c_settings.h"

void TWI_Init();
void TWI_Term();
uint8_t TWI_Start();
uint8_t TWI_Stop();
uint8_t TWI_Write(uint8_t AData);
uint8_t TWI_Read(uint8_t AAck, uint8_t *AData);

#if (EEPROM_I2C_PAGE_WRITE_SIZE!=0)
  uint16_t eepromaddress,pagesizeremaining;
#endif

uint8_t EEP_Start(uint16_t AAddress, uint8_t *AI2CAddress)
{
  uint8_t c1,c3;
  #if (EEPROM_I2C_BYTE_ADDRESSING==2)
  uint8_t c2;
  #endif
  uint16_t ui1;

  //calculate address bytes
  c3=AAddress&255;
  AAddress>>=8;
  #if (EEPROM_I2C_BYTE_ADDRESSING==2)
    c2=AAddress&255;
    AAddress>>=8;
  #endif
  c1=EEPROM_I2C_ADDRESS | (AAddress<<1);

  //wait up to EEPROM_I2C_ACK_TIMEOUT_MS milliseconds to get an start/acknowledgement 
  for (ui1=0;;ui1++)
    {
      //start
      if (!TWI_Start())
        return 0;

      //address
      if (TWI_Write(c1))      
        break;

      //stop
      if (!TWI_Stop())
        return 0;

      //return if timeout expired
      if (ui1==EEPROM_I2C_ACK_TIMEOUT_MS*25)
        return 0;

      //delay until next attempt
      Delay_us(40);
    }

  #if (EEPROM_I2C_BYTE_ADDRESSING==2)
    //address
    if (!TWI_Write(c2))
      {
        TWI_Stop();
        return 0;
      }
  #endif

  //address
  if (!TWI_Write(c3))
    {
      TWI_Stop();
      return 0;
    }

  //return i2c address if pointer is not null
  if (AI2CAddress!=NULL)
    *AI2CAddress=c1;

  //return ok
  return 1;
}

uint8_t EEP_ReadBegin(uint16_t AAddress)
{
  uint8_t c1;

  if (!EEP_Start(AAddress,&c1))
    return 0;

  //repeat start
  if (!TWI_Start())
    return 0;

  //sla_r
  if (!TWI_Write(c1 | _BV(0)))
    {
      TWI_Stop();
      return 0;
    }

  //return ok
  return 1;
}

uint8_t EEP_ReadByte(uint8_t *AByte)
{
  uint8_t c1;

  //read
  if (!TWI_Read(1,&c1))
    return 0;

  //set byte
  *AByte=c1;

  //return ok
  return 1;
}

uint8_t EEP_ReadBlock(uint16_t ALength, uint8_t *ABlock)
{
  uint8_t c1;

  //iterate length
  while (ALength--)
    {
      //read byte
      if (!EEP_ReadByte(&c1))
        return 0;

      //set byte in block
      *ABlock=c1;

      //increment block pointer
      ABlock++;
    }

  //return ok
  return 1;
}

uint8_t EEP_ReadEnd()
{
  uint8_t c1;
  
  //dummy read with nak
  if (!TWI_Read(0,&c1))
    return 0;

  //stop
  TWI_Stop();

  //return ok
  return 1;
}
uint8_t EEP_WriteBegin(uint16_t AAddress)
{
  //start
  if (!EEP_Start(AAddress,NULL))
    return 0;

  #if (EEPROM_I2C_PAGE_WRITE_SIZE!=0)
    //set eeprom address and keep track so we can start a new page write later
    eepromaddress=AAddress;

    //set page size remaining for this page
    pagesizeremaining=EEPROM_I2C_PAGE_WRITE_SIZE-(AAddress % EEPROM_I2C_PAGE_WRITE_SIZE);
  #endif

  //return ok
  return 1;
}

uint8_t EEP_WriteByte(uint8_t AByte)
{
  #if (EEPROM_I2C_PAGE_WRITE_SIZE!=0)
    if (pagesizeremaining==0)
      {
        if (!EEP_WriteEnd())
          return 0;

        if (!EEP_WriteBegin(eepromaddress))
          return 0;

        pagesizeremaining=EEPROM_I2C_PAGE_WRITE_SIZE;
      }
  #endif

  if (!TWI_Write(AByte))
    return 0;

  #if (EEPROM_I2C_PAGE_WRITE_SIZE!=0)
    //decrement page size
    pagesizeremaining--;

    //increment eeprom address
    eepromaddress++;
  #endif

  //return ok
  return 1;
}

uint8_t EEP_WriteBlock(uint16_t ALength, uint8_t *ABlock)
{
  //iterate length
  while (ALength--)
    {
      //write byte
      if (!EEP_WriteByte(*ABlock))
        return 0;

      //increment block pointer
      ABlock++;
    }

  //return ok
  return 1;
}

uint8_t EEP_WriteEnd()
{
  //stop
  TWI_Stop();

  //return ok
  return 1;
}

