#include "string.h"
#include "avr/pgmspace.h"
#include "util/crc16.h"
#include "util/atomic.h"
#include "util/twi.h"

#include "cdelay.h"
#include "ddrpin.h"

#include "structure.h"

uint8_t EEP_ReadBegin(uint16_t AAddress);
uint8_t EEP_ReadByte(uint8_t *AByte);
uint8_t EEP_ReadBlock(uint16_t ALength, uint8_t *ABlock);
uint8_t EEP_ReadEnd();

uint8_t EEP_WriteBegin(uint16_t AAddress);
uint8_t EEP_WriteByte(uint8_t AByte);
uint8_t EEP_WriteBlock(uint16_t ALength, uint8_t *ABlock);
uint8_t EEP_WriteEnd();

struct StructureType
  {
    uint16_t Address,TotalSize,StructureSize,SlotSize;
  };

#include "structure_settings.h"

struct StructureVarsType
  {
    uint16_t SlotPointer[2],WriteCount[2];
  } StructureVars[sizeof(Structures)/sizeof(struct StructureType)];

struct SlotPointerType
  {
    uint16_t SlotPointer,Crc;
  };

uint8_t StructureLoad(uint8_t AStructureNo, uint8_t *AStructure)
{
  uint8_t segment,loadedstructure,c1,c2;
  uint16_t address,slots,ui1,crc;
  struct SlotPointerType sp;
  struct StructureType st;

  //copy structure to local variable
  memcpy_P((uint8_t*)&st,(uint8_t*)&Structures[AStructureNo],sizeof(struct StructureType));

  //clear loadedstructure
  loadedstructure=0;

  //for each segment (master, backup)
  for (segment=0;segment<2;segment++)
    {
      //a 0 slot pointer means no valid slot pointer is found in flash
      StructureVars[AStructureNo].SlotPointer[segment]=0;

      //it will be incremented to 0 when written the first time
      StructureVars[AStructureNo].WriteCount[segment]=65535;

      //calculate address
      address=st.Address;
      ui1=st.TotalSize/2;
      slots=ui1/st.SlotSize;
      if (segment)
        address+=ui1;

      //read slot pointer
      if (!EEP_ReadBegin(address))
        continue;
      c2=EEP_ReadBlock(4,(uint8_t*)&sp);
      EEP_ReadEnd();
      if (!c2)
        continue;

      //validate slot pointer
      crc=0xffff;
      crc=_crc16_update(crc,*(uint8_t*)&sp.SlotPointer);
      crc=_crc16_update(crc,*((uint8_t*)&sp.SlotPointer+1));
      crc=~crc;
      crc+=0x1234;
      if (crc==sp.Crc && sp.SlotPointer && sp.SlotPointer<slots)
        {
          //set slot pointer
          StructureVars[AStructureNo].SlotPointer[segment]=sp.SlotPointer;

          //calculate slot address
          address+=st.SlotSize*sp.SlotPointer;

          //read slot
          if (!EEP_ReadBegin(address))
            continue;

          crc=0xffff;
          for (ui1=0;ui1<st.SlotSize-2;ui1++)
            {
              c2=c1;
              if (!EEP_ReadByte(&c1))
                goto eepfail;

              crc=_crc16_update(crc,c1);

              if (!loadedstructure && ui1<st.StructureSize)
                AStructure[ui1]=c1;
            }
          crc=~crc;
          crc+=0x1234;

          if (!EEP_ReadBlock(2,(uint8_t*)&ui1))
            {
              eepfail:
              EEP_ReadEnd();
              continue;
            }
          EEP_ReadEnd();

          if (crc==ui1)
            {
              //get write count
              ui1=c1<<8 | c2;

              //set write count
              StructureVars[AStructureNo].WriteCount[segment]=ui1;

              //set loaded structure
              loadedstructure=1;
            }
        }
    }

  //return loadedstructure
  return loadedstructure;
}

uint8_t StructureSave(uint8_t AStructureNo, uint8_t *AStructure)
{
  uint8_t segment,updateslotpointer,savedstructure,c1,attempt,remaponce;
  uint16_t address,ui1,crc;
  struct SlotPointerType sp,sp2;
  struct StructureType st;

  //copy structure to local variable
  memcpy_P((uint8_t*)&st,(uint8_t*)&Structures[AStructureNo],sizeof(struct StructureType));

  //clear savedstructure
  savedstructure=0;

  //clear updateslotpointer
  updateslotpointer=0;

  //clear remaponce
  remaponce=0;

  //for each segment (master, backup)
  for (segment=0;segment<2;segment++)
    {
      //if slot pointer is invalid
      if (StructureVars[AStructureNo].SlotPointer[segment]==0)
        {
          //set default slot pointer value
          StructureVars[AStructureNo].SlotPointer[segment]=1;

          //set updateslotpointer
          updateslotpointer=1;
        }

      //increment writecount
      StructureVars[AStructureNo].WriteCount[segment]++;
      if (StructureVars[AStructureNo].WriteCount[segment]>=MAX_WRITE_COUNT)
        {
          newslot:

          //clear writecount
          StructureVars[AStructureNo].WriteCount[segment]=0;

          #if (ERASE_INACTIVE_SLOTS==1)
          address=st.Address+st.SlotSize*StructureVars[AStructureNo].SlotPointer[segment];
          if (segment)
            address+=st.TotalSize/2;
          if (EEP_WriteBegin(address))
            {
              for (ui1=0;ui1<st.SlotSize;ui1++)
                if (!EEP_WriteByte(0xFF))
                  break;
              EEP_WriteEnd();
            }
          #endif

          //increment slotpointer
          StructureVars[AStructureNo].SlotPointer[segment]++;
          if (StructureVars[AStructureNo].SlotPointer[segment]>=st.TotalSize/2/st.SlotSize)
            StructureVars[AStructureNo].SlotPointer[segment]=1;

          //set updateslotpointer
          updateslotpointer=1;
        }

      //calculate address
      address=st.Address+st.SlotSize*StructureVars[AStructureNo].SlotPointer[segment];
      if (segment)
        address+=st.TotalSize/2;

      //attempt writing twice
      for (attempt=0;attempt<2;attempt++)
        {
          //write slot
          if (!EEP_WriteBegin(address))
            continue;

          //write structure data into slot
          crc=0xffff;
          for (ui1=0;ui1<st.StructureSize;ui1++)
            {
              if (!EEP_WriteByte(AStructure[ui1]))
                goto eepfail;

              crc=_crc16_update(crc,AStructure[ui1]);
            }

          //write pad data into slot (0xFF)
          for (;ui1<st.SlotSize-4;ui1++)
            {
              if (!EEP_WriteByte(0xFF))
                goto eepfail;

              crc=_crc16_update(crc,0xFF);
            }

          //write writecount into slot
          if (!EEP_WriteBlock(2,(uint8_t*)&StructureVars[AStructureNo].WriteCount[segment]))
            {
              eepfail:
              EEP_WriteEnd();
              continue;         
            }
          crc=_crc16_update(crc,*(uint8_t*)&StructureVars[AStructureNo].WriteCount[segment]);
          crc=_crc16_update(crc,*((uint8_t*)&StructureVars[AStructureNo].WriteCount[segment]+1));

          //finish crc
          crc=~crc;
          crc+=0x1234;

          //write crc16 into slot
          if (!EEP_WriteBlock(2,(uint8_t*)&crc))
            goto eepfail;
          EEP_WriteEnd();

          //verify slot
          if (!EEP_ReadBegin(address))
            continue;

          //verify structure data
          for (ui1=0;ui1<st.StructureSize;ui1++)
            {
              if (!EEP_ReadByte(&c1))
                goto eepfail2;

              if (c1!=AStructure[ui1])
                goto eepfail2;
            }

          //verify pad data (0xFF)
          for (;ui1<st.SlotSize-4;ui1++)
            {
              if (!EEP_ReadByte(&c1))
                goto eepfail2;

              if (c1!=0xFF)
                goto eepfail2;
            }

          //verify writecount
          if (!EEP_ReadBlock(2,(uint8_t*)&ui1))
            {
              eepfail2:
              EEP_ReadEnd();
              continue;
            }
          if (ui1!=StructureVars[AStructureNo].WriteCount[segment])
            goto eepfail2;

          //verify crc16
          if (!EEP_ReadBlock(2,(uint8_t*)&ui1))
            goto eepfail2;
          if (ui1!=crc)
            goto eepfail2;
          EEP_ReadEnd();

          //it is verified
          break;
        }
      if (attempt==2)
        {
          if (remaponce==0)
            {
              remaponce=1;
              goto newslot;
            }
        }
      else savedstructure=1;

      if (updateslotpointer)
        {
          //write slot pointer
          sp.SlotPointer=StructureVars[AStructureNo].SlotPointer[segment];
          crc=0xffff;
          crc=_crc16_update(crc,*(uint8_t*)&sp.SlotPointer);
          crc=_crc16_update(crc,*((uint8_t*)&sp.SlotPointer+1));
          crc=~crc;
          crc+=0x1234;
          sp.Crc=crc;

          //calculate address
          address=st.Address;
          if (segment)
            address+=st.TotalSize/2;

          for (attempt=0;attempt<2;attempt++)
            {
              if (!EEP_WriteBegin(address))
                continue;
              c1=EEP_WriteBlock(4,(uint8_t*)&sp);
              EEP_WriteEnd();
              if (!c1)
                continue;

              if (!EEP_ReadBegin(address))
                continue;
              c1=EEP_ReadBlock(4,(uint8_t*)&sp2);
              EEP_ReadEnd();
              if (!c1 || memcmp((uint8_t*)&sp,(uint8_t*)&sp2,sizeof(struct SlotPointerType))!=0)
                continue;

              //it is verified
              break;
            }
        }
    }

  //return savedstructure
  return savedstructure;
}

